@@ -13,7 +13,7 @@ char *strdup_err (const char *str, int quit)
 {
 	char *ret = strdup(str);
 	if (ret == NULL){
-		fprintf (stderr,"Can't allocate %d bytes of memory\n",strlen(str)+1);
+		depmod_error ("Can't allocate %d bytes of memory",strlen(str)+1);
 		if (quit) exit (-1);
 	}
 	return ret;
@@ -22,7 +22,7 @@ void *malloc_err (int size, int quit)
 {
 	void *ret = malloc(size);
 	if (ret == NULL){
-		fprintf (stderr,"Can't allocate %d bytes of memory\n",size);
+		depmod_error ("Can't allocate %d bytes of memory",size);
 		if (quit) exit (-1);
 	}
 	return ret;
