@@ -33,6 +33,9 @@
  * Added syslog error reporting with option "-s": Jacques Gelinas
  * Added MOD_AUTOCLEAN and option "-k" (for kerneld"): Bjorn and Jacques
  * mc68000: Andreas Schwab <schwab@issan.informatik.uni-dortmund.de>
+ * More fixes (ELF SHN_UNDEF et al) by  Andreas Schwab in February 1996
+ * Added Sparc support: Eddie C. Dost <ecd@skynet.be>
+ * Various improvements added by Hans Lermen <lermen@elserv.ffm.fgan.de>
  */
 
 #ifndef MOD_AUTOCLEAN /* defined in <linux/module.h> by the "kerneld"-patch */
@@ -65,13 +68,12 @@ static char *default_path[] = {
 	0
 };
 
-#define PATCH_STRINGS
-
 #include "insmod.h"
+#include "../Version.h"
 #define is_global(sp) (aout_flag?(sp->u.n.n_type & N_EXT) : \
-				(ELF32_ST_BIND(sp->u.e.st_info) == STB_GLOBAL))
+				(ELF32_ST_BIND(sp->u.e.st_info) != STB_LOCAL))
 #define is_undef(sp) (aout_flag?((sp->u.n.n_type & ~N_EXT) == N_UNDF) : \
-			    (ELF32_ST_TYPE(sp->u.e.st_info) == STT_NOTYPE))
+			    (sp->u.e.st_shndx == SHN_UNDEF))
 
 /* hack: sizeof(struct elfhdr) > sizeof(struct exec) */
 Elf32_Ehdr header;
@@ -90,17 +92,18 @@ struct symbol *symtab;
 char *stringtab;
 unsigned long addr;
 int verbose = 0;
+int silent_poll = 0;
 
-static int export_flag = 1; /* See comment att option handler in main() */
+static int export_flag = 1; /* See comment at option handler in main() */
 static int force_load = 0;
 static int loadmap = 0;
 static int versioned_kernel = 0;
 static struct kernel_sym nullsym;
 static struct symbol *symroot;
 static struct utsname uts_info;
+static char kd_dbmfile[40];
 
-void *
-ckalloc(size_t nbytes)
+void * ckalloc(size_t nbytes)
 {
 	void *p;
 
@@ -111,8 +114,7 @@ ckalloc(size_t nbytes)
 	return p;
 }
 
-void *
-ckrealloc(void *ptr, size_t nbytes)
+void * ckrealloc(void *ptr, size_t nbytes)
 {
         void *p;
 
@@ -145,7 +147,7 @@ static int get_kernel_syms(struct kernel_sym *buffer)
 	return syscall( __NR_get_kernel_syms, buffer);
 }
 
-void check_version(int force_load)
+static void check_version(int force_load)
 {
 	unsigned long kernel_version;
 
@@ -169,8 +171,7 @@ void check_version(int force_load)
 /*
  * unversioned kernel, versioned module
  */
-int
-m_strncmp(const char *tabentry, const char *lookfor, size_t n)
+static int m_strncmp(const char *tabentry, const char *lookfor, size_t n)
 {
 	int len = strlen(lookfor);
 	int retval;
@@ -188,8 +189,7 @@ m_strncmp(const char *tabentry, const char *lookfor, size_t n)
 /*
  * versioned kernel, unversioned module
  */
-int
-k_strncmp(const char *tabentry, const char *lookfor, size_t n)
+static int k_strncmp(const char *tabentry, const char *lookfor, size_t n)
 {
 	int len = strlen(tabentry);
 	int retval;
@@ -204,16 +204,14 @@ k_strncmp(const char *tabentry, const char *lookfor, size_t n)
 		return strcmp(tabentry, lookfor);
 }
 
-#ifdef PATCH_STRINGS
 struct strpatch {
-	int *where;
+	int where; /* offset from start */
 	int what;
 };
 static struct strpatch *stringpatches;
 static int n_stringpatches;
 
-static void
-push_string(char *string, int *patchme)
+static void push_string(char *string, int patch_offset)
 {
 	int len;
 	unsigned int string_offset = progsize;
@@ -250,14 +248,12 @@ push_string(char *string, int *patchme)
 		stringpatches = (struct strpatch *)ckrealloc(stringpatches,
 			(n_stringpatches + 1) * sizeof(struct strpatch));
 
-	(*(stringpatches + n_stringpatches)).where = patchme;
+	(*(stringpatches + n_stringpatches)).where = patch_offset;
 	(*(stringpatches + n_stringpatches)).what = string_offset;
 	++n_stringpatches;
 }
-#endif
 
-int
-main(int argc, char **argv)
+int main(int argc, char **argv)
 {
 	FILE *fp;
 	struct exec *aouthdr = (struct exec *)&header;
@@ -267,6 +263,7 @@ main(int argc, char **argv)
 	struct kernel_sym *resident_ksym_start = NULL;
 	struct mod_routines routines;
 	struct symbol *sp;
+	struct strpatch *patch;
 	unsigned long init_func, cleanup_func;
 	int (*tabcomp) (const char *, const char *, size_t) = strncmp;
 	int fatal_error;
@@ -299,7 +296,7 @@ main(int argc, char **argv)
 		p = argv[0];
 
 	if (strcmp(p, "rmmod") == 0)
-		return rmmod(argc, argv);
+		return (rmmod(argc, argv))?1:0;
 
 	if (strcmp(p, "ksyms") == 0)
 		return ksyms(argc, argv);
@@ -310,40 +307,46 @@ main(int argc, char **argv)
 		p = &(argv[1][1]);
 		while (*p) {
 			switch (*p) {
-			case 'o':
-				modname = argv[2];
-				--argc;
-				++argv;
-				break;
-
 			case 'f': /* force loading */
 				force_load = 1;
 				break;
 
-			case 'X': /* _do_ export externs */
-				export_flag = 1;
-				break;
-
-			case 'x': /* do _not_ export externs */
-				export_flag = 0;
-				break;
-
-			case 'm': /* generate loadmap */
-				loadmap = 1;
-				break;
-
 			case 'k': /* module loaded by kerneld, auto-cleanable */
 				autoclean = MOD_AUTOCLEAN;
 				break;
 
-			case 'v': /* verbose output */
-				verbose = 1;
+			case 'm': /* generate loadmap */
+				loadmap = 1;
 				break;
+
+			case 'o':
+				modname = argv[2];
+				--argc;
+				++argv;
+				break;
+
+			case 'p': /* silent poll mode */
+				silent_poll = 1;
+				break;
+
 			case 's':
 				insmod_setsyslog("insmod");
 				break;
+
+			case 'v': /* verbose output */
+				verbose = 1;
+				break;
+
 			case 'V':
-				printf("Version 1.3.57\n");
+				printf("Version " MODULES_VERSION "\n");
+				break;
+
+			case 'x': /* do _not_ export externs */
+				export_flag = 0;
+				break;
+
+			case 'X': /* _do_ export externs */
+				export_flag = 1;
 				break;
 			}
 			++p;
@@ -353,12 +356,26 @@ main(int argc, char **argv)
 	}
 
 	if (argc < 2) {
-		fputs("Usage: insmod [-f] [-x] [-o name] [-m] [-s] [-v] module "
-			"[[sym=value]...]\n", stderr);
+		fputs("Usage:\n"
+		      "insmod [-fkmopsvVxX] [-o name] module [[sym=value]...]\n"
+		      "\n"
+		      "  module     Filename of a loadable kernel module (*.o)\n"
+		      "  -f         Force loading under wrong kernel version\n"
+		      "  -k         Make module (autoclean)-able\n"
+		      "  -m         Generate loadmap (so crashes can be traced)\n"
+		      "  -o name    Set internal modulname to name\n"
+		      "  -p         Poll mode, just check if the module matches the kernel\n"
+		      "  -s         Report errors via syslog\n"
+		      "  -v         Verbose output\n"
+		      "  -V         Show version\n"
+		      "  -x         do *not* export externs\n"
+		      "  -X         *do* export externs\n"
+		      , stderr);
 		exit(2);
 	}
 
 	uname(&uts_info);
+	sprintf(kd_dbmfile, "/lib/modules/%s/persist.gdbm", uts_info.release);
 
 	filename = argv[1];
 	argv += 2;
@@ -453,11 +470,11 @@ main(int argc, char **argv)
 				; 
 			*p = '\0';
 			sprintf(spare_path, "%s/%s", path, filename);
-			if (stat(spare_path, &dummy) >= 0) 
+			if ((stat(spare_path, &dummy) >= 0) && (dummy.st_mode & S_IFREG))
 				break;
 			/* else */
 			strcat(spare_path, ".o");
-			if (stat(spare_path, &dummy) >= 0) 
+			if ((stat(spare_path, &dummy) >= 0) && (dummy.st_mode & S_IFREG))
 				break;
 		}
 		free(onepath);
@@ -497,7 +514,9 @@ main(int argc, char **argv)
 #ifdef __mc68000__
 		 header.e_machine == EM_68K
 #endif
-/* Sorry SPARC */
+#ifdef __sparc__
+		 header.e_machine == EM_SPARC
+#endif
 		) {
 			char *errstr;
 
@@ -545,7 +564,9 @@ main(int argc, char **argv)
 	 */
 	switch ((versioned_kernel << 1) + versioned_module) {
 	case 0: /* unversioned_kernel, unversioned_module */
-		check_version(force_load);
+		/* Uwe Bonnes <bon@elektron.ikp.physik.th-darmstadt.de>: */
+		if (!force_load)
+			check_version(force_load);
 		tabcomp = strncmp;
 		break;
 
@@ -589,7 +610,8 @@ main(int argc, char **argv)
 			continue;
 		}
 
-		if (defsym(tabcomp, ksym->name + (elf_kernel?0:1),
+ 		if (defsym(tabcomp, ksym->name +
+		    ((!elf_kernel && (ksym->name[0] == '_'))?1:0),
 			ksym->value, N_ABS | N_EXT,
 		/* this is safe since curr_module was initialized properly */
 			(curr_module->name[1]) ?  TRANSIENT : RESIDENT)) {
@@ -623,11 +645,12 @@ main(int argc, char **argv)
 				len = symvalue(sp);
  				if (verbose)
  				        insmod_error (
- 					"bss2 sym %08x, size =%7d: %s",
+ 					"bss2 sym %X, size =%7d: %s",
  			                bss_offset, len, symname( sp));
 			}
 			else { /* I'm not sure I understand this... */
 				sp->u.e.st_info = (STB_GLOBAL << 4)| STT_OBJECT;
+				sp->u.e.st_shndx = SHN_ABS;
 				len = (sp->u.e.st_size)?(sp->u.e.st_size):4;
 			}
  			symvalue(sp) = bss_offset;
@@ -649,7 +672,8 @@ main(int argc, char **argv)
 		}
 	}
 	if (fatal_error) {
-		insmod_error ("Failed to load module! The symbols from kernel %s don't match %s",
+		insmod_error (  "Loading failed! The module symbols "
+				"(from linux-%s) don't match your linux-%s",
 			(char*) textseg + looksym("kernel_version"),
 			uts_info.release);
 		exit(2);
@@ -697,14 +721,68 @@ main(int argc, char **argv)
 
 			do {
 				++val;
+#ifdef PERSIST
+				/*
+				 * This enables the useage of previously
+				 * saved module persistent data to update
+				 * the symbol value.
+				 *
+				 * Syntax:
+				 *
+				 *	symbol=?key:default,default...
+				 *
+				 * If no value can be found for the key,
+				 * the default values will be used.
+				 */
+				if (*val == '?') {
+#include <gdbm.h>
+					GDBM_FILE dbf;
+					datum key, data;
+					char *x;
+
+					key.dptr = val + 1;
+					if ((x = strchr(val, ':')) != NULL) {
+						*x = '\0';
+						val = ++x;
+					}
+					if ((key.dptr[0] == '\'') ||
+					    (key.dptr[0] == '"')) {
+						if ((x = strchr(key.dptr, key.dptr[0]))) {
+							*x = '\0';
+					    		key.dptr++;
+						}
+					}
+					key.dsize = strlen(key.dptr) + 1;
+
+					if ((dbf = gdbm_open(kd_dbmfile, 0,
+						    GDBM_READER, 0600, NULL))) {
+						data = gdbm_fetch(dbf, key);
+						gdbm_close(dbf);
+						if (data.dptr) {
+							/* found a value! */
+							memcpy(patchme, data.dptr, data.dsize);
+							free (data.dptr);
+							/* ignore given info */
+							break;
+						}
+					}
+					/*
+					 * No persistent data found,
+					 * use the given info
+					 */
+					if (*val == '\0')
+						break;
+				}
+#endif /* PERSIST */
 				if (*val < '0' || '9' < *val) {
-#ifdef PATCH_STRINGS
+				      /*
+				       * Textseg may change during calls to 
+				       * push_string() so we store the offset
+				       *
+				       * Timo Kokkonen <timo@cs.ualberta.ca>
+				       */
 					if (*val != ',')
-						push_string(val, patchme);
-#else
-					insmod_error ("Symbol '%s' has an illegal value: %s", param, val);
-					exit(2);
-#endif
+				     		push_string(val, symvalue(change));
 				}
 				else { /* numerical */
 					if (val[0] == '0') {
@@ -722,6 +800,13 @@ main(int argc, char **argv)
 		}
 	}
 
+	/*
+	 * If we have survived this far, and are just checking,
+	 * that's it...
+	 */
+	if (silent_poll)
+		exit(0);
+
 	/* create the module */
 	errno = 0;
 	/* make sure we have enough memory malloc'd for copy which kernel
@@ -750,19 +835,32 @@ main(int argc, char **argv)
 	else
 		relocate_elf(fp, textseg - otextseg);
 
-#ifdef PATCH_STRINGS
-	{
-		struct strpatch *p = stringpatches;
-
-		while (--n_stringpatches >= 0) {
-			*(p->where) = (int)addr + p->what;
-			++p;
-		}
-
-		free(stringpatches);
-		n_stringpatches = 0;
+	/*
+	 * Patch in any new strings from the command line
+	 */
+	for (patch = stringpatches; n_stringpatches-- > 0; ++patch) {
+		/*
+		 * Now we have to calculate the "absolute" address of the
+		 * string and update the "pointer" in the module image...
+		 *
+		 * Timo Kokkonen <timo@cs.ualberta.ca>
+		 *
+		 * Pointer increment fixes by:
+		 *   Alberto Vignani <alberto.vignani@torino.alpcom.it>
+		 *   Wolfgang Wander <wwc@rem-wwc.desy.de>
+		 */
+		*((int *)(textseg + patch->where)) = addr + patch->what;
+		/* insmod address */			 /* kernel address */
 	}
-#endif
+	if (stringpatches)
+		free(stringpatches);
+	stringpatches = (struct strpatch *)0;
+	n_stringpatches = 0;
+
+
+	/*
+	 * Handle module "administrativia"
+	 */
 	init_func += addr;
 	cleanup_func += addr;
 
@@ -855,9 +953,7 @@ main(int argc, char **argv)
 		delete_module(modname);
 		exit(1);
 	}
-#ifdef DEBUG_DISASM
-dis(textseg, codesize);
-#endif
+
 	/*
 	 * Print a loadmap so that kernel panics can be identified.
 	 * Load map option conceived by Derek Atkins <warlord@MIT.EDU>
@@ -878,10 +974,12 @@ dis(textseg, codesize);
 			case N_BSS: symtype = 'b'; break;
 			}
 		}
+#if 0
 		else
 			symtype = ' '; /* until someone does the work... */
+#endif
 
-		printf("%08lx %c %s\n",
+		insmod_printf("%X %c %s\n",
 			symvalue(sp) + addr, symtype, symname(sp));
 	}
 
@@ -891,8 +989,7 @@ dis(textseg, codesize);
 	exit(0);
 }
 
-void
-hidesym(const char *name)
+void hidesym(const char *name)
 {
 	struct symbol *sp;
 
@@ -905,8 +1002,7 @@ hidesym(const char *name)
 	}
 }
 
-int
-defsym(int (*pfi) (const char *, const char *, size_t),
+int defsym(int (*pfi) (const char *, const char *, size_t),
 	const char *name, unsigned long value, int type, int source)
 {
 	struct symbol *sp;
@@ -935,15 +1031,15 @@ defsym(int (*pfi) (const char *, const char *, size_t),
 		sp->u.n.n_type = type;
 	}
 	else { /* elf */
-		if ((ELF32_ST_BIND(sp->u.e.st_info) != STB_GLOBAL) ||
-		    (ELF32_ST_TYPE(sp->u.e.st_info) != STT_NOTYPE))
+		if ((ELF32_ST_BIND(sp->u.e.st_info) == STB_LOCAL) ||
+		    (sp->u.e.st_shndx != SHN_UNDEF))
 			return 0; /* symbol not used */
 		/* hack follows... */
 		if (type & N_ABS)
 			sp->u.e.st_info = (STB_GLOBAL << 4);
 		else
 			sp->u.e.st_info = (STB_LOCAL << 4);
-		sp->u.e.st_info |= STT_OBJECT; /* or whatever */
+		sp->u.e.st_shndx = SHN_ABS; /* or whatever */
 	}
 
 	symother(sp) |= source;
@@ -957,8 +1053,7 @@ defsym(int (*pfi) (const char *, const char *, size_t),
  * Look up an name in the symbol table.  If "add" is not null, add a
  * the entry to the table.  The table is stored as a splay tree.
  */
-struct symbol *
-findsym(const char *key, struct symbol *add,
+struct symbol * findsym(const char *key, struct symbol *add,
 	int (*pfi) (const char *, const char *, size_t))
 {
 	struct symbol *left, *right;
@@ -1058,8 +1153,7 @@ one_level_only:
 }
 
 
-unsigned long
-looksym(const char *name)
+unsigned long looksym(const char *name)
 {
 	struct symbol *sp;
 
@@ -1093,12 +1187,11 @@ struct mod {
 	struct ref *ref;
 };
 
-struct mod *loaded;
-int current = -1;
+static struct mod *loaded;
+static int current = -1;
 
 /* build the references as shown in /proc/ksyms */
-void
-get_stacks()
+static void get_stacks()
 {
 	FILE *fp;
 	struct ref *rp;
@@ -1161,29 +1254,28 @@ get_stacks()
 	fclose(fp);
 }
 
-int
-rmmod(int argc, char **argv)
+int rmmod(int argc, char **argv)
 {
 	struct ref *rp;
 	int i, j;
+	int errors = 0;
 	int count;
 	char **list;
 	char *p;
 
 	if (argc == 1) {
 		fprintf(stderr, "usage: rmmod [-r] [-s] module ...\n");
-		exit(1);
+		return 1;
 	}
 	/* else */
 
 	if (strcmp(argv[1], "-r") != 0) {
 		if (strcmp(argv[1], "-a") == 0) {
 			/* delete all unused modules and stacks */
-			delete_module(NULL);
-			exit(0);
+			return (delete_module(NULL))?1:0;
 		}
 		/* else */
-		if (strcmp(argv[1],"-s")==0){
+		if (strcmp(argv[1], "-s") == 0){
 			insmod_setsyslog("rmmod");
 			--argc;
 			++argv;
@@ -1194,12 +1286,13 @@ rmmod(int argc, char **argv)
 				 (strcmp(p, ".mod") == 0))) 
 					*p = '\0';
 			if (delete_module(argv[1]) < 0) {
+				++errors;
 				perror(argv[1]);
 			}
 			--argc;
 			++argv;
 		}
-		exit(0);
+		return errors;
 	}
 	/* else recursive removal */
 
@@ -1207,7 +1300,7 @@ rmmod(int argc, char **argv)
 	list = &(argv[2]);
 	if (count <= 0) {
 		insmod_error ("usage: rmmod [-r] module ...");
-		exit(1);
+		return 1;
 	}
 
 	get_stacks();
@@ -1219,8 +1312,10 @@ rmmod(int argc, char **argv)
 				break;
 			}
 		}
-		if (j > current)
-			insmod_error ("module '%s' not loaded", list[i]);
+		if (j > current) {
+			insmod_error("module '%s' not loaded", list[i]);
+			++errors;
+		}
 	}
 
 	for (i = 0; i <= current; ++i) {
@@ -1247,18 +1342,19 @@ rmmod(int argc, char **argv)
 		case CAN_REMOVE:
 		case WANT_TO_REMOVE | CAN_REMOVE:
 			if (delete_module(loaded[i].name) < 0) {
+				++errors;
 				perror(loaded[i].name);
 			}
 			break;
 
 		case WANT_TO_REMOVE:
-			insmod_error ("module '%s' is in use!",
-				loaded[i].name);
+			insmod_error("module '%s' is in use!", loaded[i].name);
+			++errors;
 			break;
 		}
 	}
 
-	return 0;
+	return errors;
 }
 
 
@@ -1271,8 +1367,7 @@ rmmod(int argc, char **argv)
  * See the file COPYING for your rights (GNU GPL)
  */
 
-int
-ksyms(int argc, char **argv)
+int ksyms(int argc, char **argv)
 {
 	struct kernel_sym *ksymtab = NULL;
 	struct kernel_sym *ksym = NULL;
@@ -1312,7 +1407,7 @@ ksyms(int argc, char **argv)
 
 	if (argc < 1) {
 		fputs("Usage: ksyms [-a] [-h]\n", stderr);
-		exit(2);
+		return 2;
 	}
 
 	/* get the size of the current kernel symbol table */
@@ -1320,15 +1415,15 @@ ksyms(int argc, char **argv)
 
 	if (nksyms < 0) {
 		insmod_error ("get_kernel_sys failed: Cannot find Kernel symbols!");
-		exit(2);
+		return 2;
 	}
 
 	if (nksyms) {
 		ksymtab = (struct kernel_sym *) ckalloc(nksyms * sizeof *ksymtab);
 		/* NOTE!!! The order of the symbols is important */
 		if (get_kernel_syms(ksymtab) != nksyms) {
-			insmod_error ("Kernel symbol problem");
-			exit(2);
+			insmod_error("Kernel symbol problem");
+			return 2;
 		}
 	}
 
@@ -1353,20 +1448,20 @@ ksyms(int argc, char **argv)
 		(lseek(kmem, (off_t)ksym->value, SEEK_SET) > 0) &&
 		(read(kmem, (char *)&module_struct, sizeof(struct module)) ==
 		sizeof(struct module))) {
-					printf("%08lx --- (%dk) ---\t[%s]\n",
+					printf("%lX --- (%dk) ---\t[%s]\n",
 					(long)module_struct.addr,
 					module_struct.size * 4,
 					module_name);
 				}
 				else {
 					perror("/dev/kmem");
-					exit(2);
+					return 2;
 				}
 			}
 			continue;
 		}
 
-		printf("%08lx %s", ksym->value, ksym->name);
+		printf("%lX %s", ksym->value, ksym->name);
 		if (*module_name)
 			printf("\t[%s]", module_name);
 		printf("\n");
