@@ -15,8 +15,6 @@
  * Support for versioned kernels and symbols: Bjorn Ekwall in December 1994
  */
 
-/*#define SHAME_ON_YOU*/
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -65,21 +63,31 @@ struct symbol {
 /* hack: sizeof(struct elfhdr) > sizeof(struct exec) */
 extern Elf32_Ehdr header;
 extern int aout_flag;
+extern int elf_kernel;
+extern int verbose;
 extern char *load_aout(FILE *);
-extern void relocate_aout(FILE *);
+extern void relocate_aout(FILE *, long);
 extern char *load_elf(FILE *);
-extern void relocate_elf(FILE *);
+extern void relocate_elf(FILE *, long);
 
 void *ckalloc(size_t);
 void hidesym(const char *);
-int defsym(const char *, unsigned long, int, int);
+int defsym(int (*pfi)(const char *, const char *, size_t), const char *, unsigned long, int, int);
 struct symbol *findsym(const char *, struct symbol *, int (*pfi)(const char *, const char *, size_t));
 unsigned long looksym(const char *);
 
-extern int codesize;
-extern int progsize;
+extern size_t codesize;
+extern size_t progsize;
+extern size_t bss1size;
+extern size_t bss2size;
 extern char *textseg;
 extern int nsymbols;
 struct symbol *symtab;
 extern char *stringtab;
 extern unsigned long addr;
+
+/* error.c 29/05/95 21.43.42 */
+void insmod_setsyslog (const char *name);
+void insmod_error (const char *ctl, ...);
+void insmod_debug (const char *ctl, ...);
+void insmod_printf (const char *ctl, ...);
