@@ -11,6 +11,10 @@
  * 	James Bottomley <J.E.J.Bottomley@damtp.cambridge.ac.uk>
  * mc68000 additions:
  *	Andreas Schwab <schwab@issan.informatik.uni-dortmund.de>
+ *
+ * More fixes (ELF SHN_UNDEF et al) by  Andreas Schwab in February 1996
+ * sparc additions:
+ *	Eddie C. Dost <ecd@skynet.be>
  */
 #include "insmod.h"
 
@@ -94,7 +98,7 @@ elf_relocate(unsigned int loadaddr, void *secpnt, int n_rel)
 	int got_addr = (int)got - (int)textseg + addr;
 	Elf32_Rel *rpnt = (Elf32_Rel *) secpnt;
 #endif
-#ifdef __mc68000__
+#if defined(__mc68000__) || defined(__sparc__)
 	Elf32_Rela *rpnt = (Elf32_Rela *) secpnt;
 #endif
 
@@ -127,7 +131,9 @@ elf_relocate(unsigned int loadaddr, void *secpnt, int n_rel)
 			/* kludge done at loading:
 			 * st_value = offset from image start!
 			 */
-			if (sp->u.e.st_shndx && secref[sp->u.e.st_shndx]) {
+			if (sp->u.e.st_shndx &&
+			    sp->u.e.st_shndx < SHN_LORESERVE &&
+			    secref[sp->u.e.st_shndx]) {
 				symbol_addr = (unsigned int)sp->u.e.st_value +
 						addr;
 			}
@@ -222,6 +228,111 @@ elf_relocate(unsigned int loadaddr, void *secpnt, int n_rel)
 					+ rpnt->r_addend);
 			break;
 #endif
+#ifdef __sparc__
+		case R_SPARC_NONE:
+			break;
+
+		case R_SPARC_8:
+			*reloc_addr &= ~(0xff);
+			*reloc_addr |= (symbol_addr + rpnt->r_addend) & 0xff;
+			break;
+
+		case R_SPARC_16:
+			*reloc_addr &= ~(0xffff);
+			*reloc_addr |= (symbol_addr + rpnt->r_addend) & 0xffff;
+			break;
+
+		case R_SPARC_32:
+			*reloc_addr = symbol_addr + rpnt->r_addend;
+			break;
+
+		case R_SPARC_DISP8:
+			insmod_error("unhandled reloc SPARC_DISP8, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_DISP16:
+			insmod_error("unhandled reloc SPARC_DISP16, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_DISP32:
+			insmod_error("unhandled reloc SPARC_DISP32, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+
+		case R_SPARC_WDISP30:
+			*reloc_addr &= ~(0x3fffffff);
+			*reloc_addr |= (symbol_addr - real_reloc_addr) >> 2;
+			break;
+
+		case R_SPARC_WDISP22:
+			insmod_error("unhandled reloc SPARC_WDISP22, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+
+		case R_SPARC_HI22:
+			*reloc_addr &= ~(0x3fffff);
+			*reloc_addr |= (symbol_addr + rpnt->r_addend) >> 10;
+			break;
+
+		case R_SPARC_22:
+			insmod_error("unhandled reloc SPARC_22, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_13:
+			insmod_error("unhandled reloc SPARC_13, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+
+		case R_SPARC_LO10:
+			*reloc_addr &= ~(0x3ff);
+			*reloc_addr |= (symbol_addr + rpnt->r_addend) & 0x3ff;
+			break;
+
+		case R_SPARC_GOT10:
+			insmod_error("unhandled reloc SPARC_GOT10, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_GOT13:
+			insmod_error("unhandled reloc SPARC_GOT13, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_GOT22:
+			insmod_error("unhandled reloc SPARC_GOT22, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_PC10:
+			insmod_error("unhandled reloc SPARC_PC10, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_PC22:
+			insmod_error("unhandled reloc SPARC_PC22, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_WPLT30:
+			insmod_error("unhandled reloc SPARC_WPLT30, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_COPY:
+			insmod_error("unhandled reloc SPARC_COPY, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_GLOB_DAT:
+			insmod_error("unhandled reloc SPARC_GLOB_DAT, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_JMP_SLOT:
+			insmod_error("unhandled reloc SPARC_JMP_SLOT, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_RELATIVE:
+			insmod_error("unhandled reloc SPARC_RELATIVE, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+		case R_SPARC_UA32:
+			insmod_error("unhandled reloc SPARC_UA32, ");
+			insmod_error("mail to ecd@skynet.be\n");
+			exit (1);
+#endif
 
 		default:
 			insmod_error ("Unable to handle reloc type %d",
@@ -263,7 +374,7 @@ relocate_elf(FILE *fp, long offset)
 			elf_relocate(loadaddr, (void *) secref[i], n_rel);
 		}
 #endif
-#ifdef __mc68000__
+#if defined(__mc68000__) || defined(__sparc__)
 		if (spnt->sh_type == SHT_RELA) {
 			loadaddr = (unsigned int)secref[spnt->sh_info];
 			n_rel= sections[i].sh_size / sections[i].sh_entsize;
@@ -278,7 +389,9 @@ relocate_elf(FILE *fp, long offset)
 		int symbol_addr;
 
 		sp = (struct symbol *)(*next_got);
-		if (sp->u.e.st_shndx && secref[sp->u.e.st_shndx]) {
+		if (sp->u.e.st_shndx &&
+		    sp->u.e.st_shndx < SHN_LORESERVE &&
+		    secref[sp->u.e.st_shndx]) {
 			symbol_addr = (unsigned int)sp->u.e.st_value + addr;
 			}
 		else {
@@ -415,7 +528,7 @@ load_elf(FILE *fp)
 #endif
 
 		case SHT_RELA:
-#ifdef __mc68000__
+#if defined(__mc68000__) || defined(__sparc__)
 			secref[i] = (char *)ckalloc(spnt->sh_size);
 
 			fseek(fp, spnt->sh_offset, SEEK_SET);
@@ -430,7 +543,7 @@ load_elf(FILE *fp)
 		case SHT_HASH: return "can't handle section HASH";
 		case SHT_DYNAMIC: return "can't handle section DYNAMIC";
 		case SHT_SHLIB: return "can't handle section SHLIB";
-		case SHT_DYNSYM: return "can't handle section DYMSYM";
+		case SHT_DYNSYM: return "can't handle section DYNSYM";
 		case SHT_NUM: return "can't handle section NUM";
 		default: return "can't handle section > 12";
 		}
@@ -456,10 +569,13 @@ load_elf(FILE *fp)
 	}
 
 	for (n = nsymbols, sp = symtab ; --n >= 0 ; sp++) {
-		/* look up name and add sp to binary tree */
+		/* Hmmm... */
+		if ((sp->u.e.st_info == 0) && (sp->u.e.st_shndx != SHN_UNDEF))
+			sp->u.e.st_info = STT_OBJECT; /* as good as any... */
+		/* look up name and add sp to splay tree */
 		findsym(stringtab + sp->u.e.st_name, sp, strncmp);
-		if ((ELF32_ST_BIND(sp->u.e.st_info) == STB_GLOBAL) &&
-			(ELF32_ST_TYPE(sp->u.e.st_info) != STT_NOTYPE))
+		if ((ELF32_ST_BIND(sp->u.e.st_info) != STB_LOCAL) &&
+		    (sp->u.e.st_shndx != SHN_UNDEF))
 			symother(sp) = DEF_BY_MODULE; /* abuse: mark extdef */
 		else {
 			symother(sp) = 0; /* abuse: mark extref */
@@ -469,8 +585,9 @@ load_elf(FILE *fp)
 		 * the start of the newly build image.
 		 * I rely on the secref mapping to be correct...
 		 */
-		if ( (sp->u.e.st_shndx < header.e_shnum) && /* always >= 0 ! */
-			secref[sp->u.e.st_shndx]) {
+		if ((sp->u.e.st_shndx != SHN_UNDEF) &&
+		    (sp->u.e.st_shndx < header.e_shnum) && /* always >= 0 ! */
+		    secref[sp->u.e.st_shndx]) {
 
 			(unsigned int)sp->u.e.st_value +=
 				(int)(secref[sp->u.e.st_shndx]) -
